#include "mns.h"

#define USAGE {printf("usage: %s <source> <dest> <port>\n", argv[0]);exit(0);}

#define OSCFGFILE "fingerprints.conf"

#define SERVICES 24
#define ACK_RAND 200
#define PACKETS 8
#define PORT_CLOSED 1

#define TH_XXX 0x40
#define TH_YYY 0x80

/* local structures */
struct fingerprint {
   u_short set:2,
           seq:2,
           ackset:3,
           urg:1;
   u_long ack;
   u_short win;
   u_short flag;
};

struct servicetype {
   char *name;
   int port;
};

typedef struct {
   u_long seq;
   u_long ack;
} hdrinfo;

/* function prototypes */
void rawsockets(int *, int *);
char *checkosmain(int, int, u_short);
void sendtcp(int, hdrinfo *, u_short);
u_long gettcp(int, struct in_addr, struct tcphdr *);

/* global variables */
time_t timedelay;
SAin sa_from, sa_dest;

void main(int argc, char *argv[]) {
   struct tcphdr tcp;
   
   int sd_send, sd_recv, port;
   char *buffer = NULL;
   int i, pcount;
   char ham_asc[2];

   setvbuf(stdout, NULL, _IONBF, 0);

   rawsockets (&sd_send, &sd_recv);

   sa_from.sin_family = sa_dest.sin_family = AF_INET;

   if(argc < 4) USAGE;
   if(!ipv4_rslvhost(argv[1], &sa_from)) USAGE;
   if(!ipv4_rslvhost(argv[2], &sa_dest)) USAGE;
   if((port = atoi(argv[3])) < 1) USAGE;

   buffer = checkosmain(sd_send, sd_recv, port);
   
   if (buffer != NULL) {
      ham_asc[0] = buffer[0];
      ham_asc[1] = '\0';
      printf("#\tTCP Sequence Prediction Difficulty -> %s ", ham_asc);
      switch (ham_asc[0]) {
         case '0': puts("( indeterminate )"); break;
         case '1': puts("( easy )"         ); break;
	 case '2': puts("( easy )"         ); break;
	 case '3': puts("( moderate )"     ); break;
	 case '4': puts("( moderate )"     ); break;
	 case '5': puts("( hard )"         ); break;
	 case '6': puts("( infeaseable )"  ); break;
	 case '7': puts("( impossible )"   ); break;
	 case '8': puts("( impossible )"   ); break;
	 default : puts("( hamming error )"); break;
      }
      printf("#\tOperating System -> %s\n#\n", &buffer[1]);
   }
}

char *checkosmain (int sd_send, int sd_recv, u_short dport) {
  FILE *fs_cfg;
  char *osname;
  struct fingerprint pak[PACKETS];
  struct tcphdr tcp;
  hdrinfo seq_ack;
  u_short start, s;
  u_long myseq[PACKETS];
  u_long seq_mask = 0, last_seq = 0;
  time_t timeout;
  char line[60];
  int n, i, ham;

  for (n = 0; n < PACKETS; n++)
    bzero (&pak[n], sizeof (struct fingerprint));

  srand (time (NULL) & 0x0000ffff);
  start = s = (u_short) getpid () + 5000 + (rand () % 5000);
  seq_ack.ack = 0;
  sa_dest.sin_port = htons(dport);

#define sendtcppacket(a, b)\
   {seq_ack.seq = myseq[a] = rand();\
    sa_from.sin_port = htons(s + a);\
    sendtcp(sd_send, &seq_ack, b);\
    usleep(10);}

  sendtcppacket (0, TH_SYN);
  sendtcppacket (1, TH_SYN | TH_ACK);
  sendtcppacket (2, TH_FIN);
  sendtcppacket (3, TH_FIN | TH_ACK);
  sendtcppacket (4, TH_SYN | TH_FIN);
  sendtcppacket (5, TH_PUSH);
  sendtcppacket (6, TH_SYN | TH_XXX | TH_YYY);
  sendtcppacket (7, TH_SYN);	/* this extra syn is used for hamming distances
				 * an even number of responses are needed */
  timeout = time(NULL) + TIMEOUT;
  n = i = 0;

  do {
     if (gettcp (sd_recv, sa_dest.sin_addr, &tcp)) {
        if (sa_dest.sin_port == tcp.th_sport) {
           if (ntohs (tcp.th_dport) - start < PACKETS) {
              n = ntohs(tcp.th_dport) - start;
              if ((n < 0) || (n >= PACKETS))
                 continue;
              if (pak[n].set)
                 continue;
              pak[n].seq = tcp.th_seq != 0;
              if (pak[n].seq) {
                 seq_mask ^= tcp.th_seq;
                 last_seq = tcp.th_seq;
                 i++;
              }
              pak[n].ack = ntohl (tcp.th_ack);
              pak[n].ackset = tcp.th_ack != 0;
              if (pak[n].ackset && (myseq[n] > pak[n].ack))
                 pak[n].ackset = 2;
              if (pak[n].ackset)
                 pak[n].ack = (myseq[n] < pak[n].ack) ?
                 (pak[n].ack - myseq[n]) :
                 (myseq[n] - pak[n].ack);
              if (pak[n].ack > ACK_RAND) {
                 pak[n].ackset = 3;
                 pak[n].ack = 0;
              }
              pak[n].win = ntohs (tcp.th_win);
              pak[n].flag = tcp.th_flags;
	      pak[n].set = 1;
              pak[n].urg = tcp.th_urp != 0;
           }
        }
     }
     usleep (10);
  }
  while ((timeout > time(NULL)) && (n != PACKETS - 1));

  /* display new OS fingerprint */
  printf ("\n* New OS Fingerprint\n");
  for (n = 0; n < PACKETS - 1; n++)
    {
      if (pak[n].set)
	{
	  printf ("%i %i ", n, pak[n].seq);
	  switch (pak[n].ackset)
	    {
	    case 0:
	      putchar ('N');
	      break;
	    case 1:
	      putchar ('+');
	      printf ("%li", pak[n].ack);
	      break;
	    case 2:
	      putchar ('-');
	      printf ("%li", pak[n].ack);
	      break;
	    case 3:
	      putchar ('R');
	      break;
	    }

	  printf (" %04x ", pak[n].win);
	  if (pak[n].flag & TH_SYN)
	    putchar ('S');
	  if (pak[n].flag & TH_RST)
	    putchar ('R');
	  if (pak[n].flag & TH_ACK)
	    putchar ('A');
	  if (pak[n].flag & TH_FIN)
	    putchar ('F');
	  if (pak[n].flag & TH_PUSH)
	    putchar ('P');
	  if (pak[n].flag & TH_XXX)
	    putchar ('X');
	  if (pak[n].flag & TH_YYY)
	    putchar ('Y');
	  if (pak[n].urg)
	    putchar ('U');
	  putchar ('\n');
	}
      else
	printf ("%i - - - -\n", n);
    }
  putchar ('\n');


  
/* calculate hamming distance */
  if (i % 2)
    seq_mask ^= last_seq;
  ham = hamming (seq_mask);
  if (ham > 8)
    ham = 16 - ham;
/* display fingerprint */


  /* parse fingerprints */
  if ((fs_cfg = fopen (OSCFGFILE, "r")) != NULL)
    {
      struct fingerprint parse;
      int flag1 = 0, found = 0, total = 0;
      char *token;

      osname = malloc (60);

      while (fgets (line, 60, fs_cfg))
	{
	  if (line[0] == '#')
	    continue;		/* COMMENT LINE */
	  if (line[0] == '\n')
	    {			/* BLANK LINE */
	      if (flag1 && (found == total))
		{
		  fclose (fs_cfg);
		  osname[0] = (char) ham + '0';
		  return osname;
		};
	    }
	  if (line[0] == '*')
	    {			/* OS FINGERPRINT HEADER */
	      if (strlen (line) > 3)
		{
		  strcpy (osname, &line[1]);
		  osname[strlen (osname) - 1] = '\0';
		}
	      flag1 = 1;
	      total = found = 0;
	      continue;
	    }

	  /* line number */
	  token = strtok (line, " ");
	  if (isdigit (token[0]))
	    n = atoi (token);
	  else
	    {
	      flag1 = total = found = 0;
	      continue;
	    }

/*-- seq --*/
	  if ((token = strtok (NULL, " ")) != NULL)
	    parse.seq = atoi (token);
	  parse.set = isdigit (token[0]) ? 1 : 0;

/*-- ack --*/
	  if ((token = strtok (NULL, " ")) != NULL)
	    {
	      switch (token[0])
		{
		case '+':
		  parse.ackset = 1;
		  parse.ack = atol (&token[1]);
		  break;
		case '-':
		  parse.ackset = 2;
		  parse.ack = atol (&token[1]);
		  break;
		case 'R':
		  parse.ackset = 3;
		  parse.ack = 0;
		  break;
		case 'N':
		default:
		  parse.ackset = 0;
		  parse.ack = 0;
		  break;
		}
	    }

/*-- win --*/
	  if ((token = strtok (NULL, " ")) != NULL)
	    parse.win = strtol (token, NULL, 16);

/*-- flags --*/
	  if ((token = strtok (NULL, " \n")) != NULL)
	    {
	      parse.flag = 0;
	      if (strchr (token, 'S'))
		parse.flag |= TH_SYN;
	      if (strchr (token, 'R'))
		parse.flag |= TH_RST;
	      if (strchr (token, 'A'))
		parse.flag |= TH_ACK;
	      if (strchr (token, 'F'))
		parse.flag |= TH_FIN;
	      if (strchr (token, 'P'))
		parse.flag |= TH_PUSH;
	      if (strchr (token, 'X'))
		parse.flag |= TH_XXX;
	      if (strchr (token, 'Y'))
		parse.flag |= TH_YYY;
	      if (strchr (token, 'U'))
		parse.urg = 1;
	      else
		parse.urg = 0;
	    }

	  total++;

	  /* test for no response */
	  if (!pak[n].set)
	    {
	      if (parse.set)
		total = found = flag1 = 0;
	      else
		found++;
	      continue;
	    }

	  if ((parse.ackset == pak[n].ackset) &&
	      (parse.ack == pak[n].ack) &&
	      (parse.seq == pak[n].seq) &&
	      (parse.win == pak[n].win) &&
	      (parse.flag == pak[n].flag) &&
	      (parse.urg == pak[n].urg))
	    {
	      found++;
	      continue;
	    }
	  else
	    {
	      total = found = flag1 = 0;
	      continue;
	    }
	}

      fseek (fs_cfg, 0L, SEEK_END);
      fclose (fs_cfg);
      if (osname != NULL)
	free (osname);

      return NULL;
    }

  fprintf (stderr, "ERROR: cannot read from %s\n", OSCFGFILE);
  return NULL;
}
/* Sends a TCP packet */
void sendtcp (int sd_send, hdrinfo * routing, u_short flags) {
  struct tcphdr tcp;

  tcp.th_sport = sa_from.sin_port;
  tcp.th_dport = sa_dest.sin_port;
  tcp.th_seq = htonl (routing->seq);
  tcp.th_ack = 0;
#ifdef X2_OFF
  tcp.th_x2_off = 0x50;
#else
  tcp.th_x2 = 0;
  tcp.th_off = 5;
#endif
  tcp.th_flags = flags;
  tcp.th_win = htons (0x1000 | (rand () % 0x1000));
  tcp.th_urp = 0;
  tcp.th_sum = 0;

/*-- TCP Checksum --*/
#ifdef SOLARIS_CKSUM_BUG
  tcp.th_sum = sizeof (struct tcphdr);
#else
  tcp.th_sum = tcp_cksum((char *)&tcp, sizeof(struct tcphdr),
			 sa_from.sin_addr, sa_dest.sin_addr);
#endif /* SOLARIS_CKSUM_BUG */

  if (sendto(sd_send, &tcp, sizeof(tcp), 0, (SA *)&sa_dest, sizeof(SAin)) < 0)
     perror ("sending message");
}

#define PACKETLEN 0xffffL
#define tcphd ((struct tcphdr *)&packet[20])
#define iphd  ((struct ip *)packet)

u_long gettcp(int sd_recv, struct in_addr targetip, struct tcphdr *dtcp) {
   char packet[PACKETLEN];
   int i;
   
   if (!recv(sd_recv, packet, PACKETLEN, 0) < 0) return 0;
   if (memcmp(&targetip, &iphd->ip_src, 4)) return 0;

   memcpy(dtcp, tcphd, sizeof(struct tcphdr));

   return iphd->ip_dst.s_addr;
}

void rawsockets (int *sd_send, int *sd_recv) {
   int on = 1;
   int rsflags;

/*-- SEND RAW socket --*/
  if ((*sd_send = socket (AF_INET, SOCK_RAW, IPPROTO_TCP)) < 0)
    {
      perror ("opening raw send socket");
      exit (EXIT_FAILURE);
    }

/*-- READ RAW socket --*/
  if ((*sd_recv = socket (AF_INET, SOCK_RAW, IPPROTO_TCP)) < 0)
    {
      perror ("opening raw read socket");
      exit (EXIT_FAILURE);
    }
  if ((rsflags = fcntl (*sd_recv, F_GETFL)) == -1)
    {
      perror ("fcntl(sd_recv,F_GETFL)");
      exit (EXIT_FAILURE);
    }
  if (fcntl (*sd_recv, F_SETFL, rsflags | O_NONBLOCK) == -1)
    {
      perror ("fcntl(sd_recv,F_SETFL)");
      exit (EXIT_FAILURE);
    }
}

